"""
This module contains various helper functions and classes.
"""
import trio
import anyio
import sys
import os
import re
import asyncclick as click

import attr
import outcome
import collections.abc
import simpleeval
import ast as _ast

from getpass import getpass
from collections import deque
from collections.abc import Mapping
from copy import deepcopy
from types import ModuleType
from typing import Union, Dict, Optional
from ssl import SSLContext
from functools import partial, total_ordering
from math import log10
from sniffio import current_async_library

import ruamel.yaml as yaml

SafeRepresenter = yaml.representer.SafeRepresenter
SafeConstructor = yaml.constructor.SafeConstructor
Emitter = yaml.emitter.Emitter

NoneType = type(None)

from .exceptions import CancelledError

import logging

logger = logging.getLogger(__name__)


try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager


def singleton(cls):
    return cls()


def yload(stream, multi=False):
    y = yaml.YAML(typ="safe")
    if multi:
        return y.load_all(stream)
    else:
        return y.load(stream)


def yprint(data, stream=sys.stdout, compact=False):
    """
    Standard code to write a YAML record.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible. default False.
    """
    if isinstance(data, (int, float)):
        print(data, file=stream)
    elif isinstance(data, (str, bytes)):
        print(repr(data), file=stream)
    #   elif isinstance(data, bytes):
    #       os.write(sys.stdout.fileno(), data)
    else:
        y = yaml.YAML(typ="safe")
        y.default_flow_style = compact
        y.dump(data, stream=stream)


def yformat(data, compact=None):
    """
    Return ``data`` as a multi-line YAML string.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible. default False.
    """
    from io import StringIO

    s = StringIO()
    yprint(data, compact=compact, stream=s)
    return s.getvalue()


_expect_node = Emitter.expect_node


def expect_node(self, *a, **kw):
    _expect_node(self, *a, **kw)
    self.root_context = False


Emitter.expect_node = _expect_node


class TimeOnlyFormatter(logging.Formatter):
    default_time_format = "%H:%M:%S"
    default_msec_format = "%s.%03d"


class NotGiven:
    """Placeholder value for 'no data' or 'deleted'."""

    def __new__(cls):
        return cls

    def __getstate__(self):
        raise ValueError("You may not serialize this object")

    def __repr__(self):
        return "‹NotGiven›"

    def __str__(self):
        return "NotGiven"


def combine_dict(*d, cls=dict, deep=False) -> dict:
    """
    Returns a dict with all keys+values of all dict arguments.
    The first found value wins.

    This recurses if values are dicts.

    Args:
      cls (type): a class to instantiate the result with. Default: dict.
        Often used: :class:`attrdict`.
      deep (bool): if set, always copy.
    """
    res = cls()
    keys = {}
    if not len(d):
        return res

    if len(d) == 1 and deep and not isinstance(d[0], Mapping):
        if deep and isinstance(d[0], (list, tuple)):
            return deepcopy(d[0])
        else:
            return d[0]

    for kv in d:
        for k, v in kv.items():
            if k not in keys:
                keys[k] = []
            keys[k].append(v)

    for k, v in keys.items():
        if v[0] is NotGiven:
            res.pop(k, None)
        elif len(v) == 1 and not deep:
            res[k] = v[0]
        elif not isinstance(v[0], Mapping):
            for vv in v[1:]:
                assert vv is NotGiven or not isinstance(vv, Mapping)
            if deep and isinstance(v[0], (list, tuple)):
                res[k] = deepcopy(v[0])
            else:
                res[k] = v[0]
        else:
            res[k] = combine_dict(*v, cls=cls)

    return res


def drop_dict(data: dict, drop: tuple) -> dict:
    data = data.copy()
    for d in drop:
        vv = data
        if isinstance(d, tuple):
            for dd in d[:-1]:
                vv = vv[dd] = vv[dd].copy()
            d = d[-1]
        del vv[d]
    return data


class attrdict(dict):
    """A dictionary which can be accessed via attributes, for convenience.

    This also supports updating path accessors.
    """

    def __getattr__(self, a):
        if a.startswith("_"):
            return object.__getattribute__(self, a)
        try:
            return self[a]
        except KeyError:
            raise AttributeError(a) from None

    def __setattr__(self, a, b):
        if a.startswith("_"):
            super(attrdict, self).__setattr__(a, b)
        else:
            self[a] = b

    def __delattr__(self, a):
        try:
            del self[a]
        except KeyError:
            raise AttributeError(a) from None

    def _get(self, path, skip_empty=True, default=NotGiven):
        """
        Get a node's value and access the dict items beneath it.
        """
        if isinstance(path, str):
            raise ValueError(f"Must be a Path/list, not {path!r}")
        val = self
        for p in path:
            if val is None:
                return None
            if skip_empty and not p:
                continue
            val = val.get(p, NotGiven)
            if val is NotGiven:
                if default is NotGiven:
                    raise KeyError(path)
                return default
        return val

    def _update(self, path, value=None, skip_empty=True):
        """
        Set some sub-item's value, possibly merging dicts.
        Items set to 'NotGiven' are deleted.

        Returns the new value. Modified (sub)dicts will be copied.
        """
        if isinstance(path, str):
            raise ValueError(f"Must be a Path/list, not {path!r}")
        if skip_empty:
            path = [p for p in path if p]
        val = type(self)(**self)
        v = val
        if not path:
            if isinstance(value, Mapping):
                return combine_dict(value, val, cls=type(self))
            else:
                return value

        for p in path[:-1]:
            try:
                w = v[p]
            except KeyError:
                w = type(v)()
            else:
                # copy
                w = type(w)(w)
            v[p] = w
            v = w
        px = path[-1]
        if value is NotGiven:
            v.pop(px, None)
        elif not isinstance(value, Mapping):
            v[px] = value
        elif px in v:
            v[px] = combine_dict(value, v[px], cls=type(self))
        else:
            v[px] = value

        return val

    def _delete(self, path, skip_empty=True):
        """
        Remove some sub-item's value, possibly removing now-empty intermediate
        dicts.

        Returns the new value. Modified (sub)dicts will be copied.
        """
        if isinstance(path, str):
            raise ValueError(f"Must be a Path/list, not {path!r}")
        if skip_empty:
            path = [p for p in path if p]
        val = type(self)(**self)
        v = val
        vc = []
        for p in path[:-1]:
            vc.append(v)
            try:
                w = v[p]
            except KeyError:
                return self
            w = type(w)(**w)
            v[p] = w
            v = w
        vc.append(v)
        while path:
            v = vc.pop()
            del v[path.pop()]
            if v:
                break
        return val


SafeRepresenter.add_representer(attrdict, SafeRepresenter.represent_dict)


def str_presenter(dumper, data):
    if "\n" in data:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


SafeRepresenter.add_representer(str, str_presenter)


def count(it):
    n = 0
    for _ in it:
        n += 1
    return n


async def acount(it):
    n = 0
    async for _ in it:  # noqa: F841
        n += 1
    return n


class _MsgRW:
    """
    Common base class for :class:`MsgReader` and :class:`MsgWriter`.
    """

    _mode = None

    def __init__(self, path=None, stream=None):
        if (path is None) == (stream is None):
            raise RuntimeError("You need to specify either path or stream")
        self.path = path
        self.stream = stream

    async def __aenter__(self):
        if self.path is not None:
            try:
                self.stream = await anyio.aopen(self.path, self._mode)
            except AttributeError:
                self.stream = await anyio.open_file(self.path, self._mode)
        return self

    async def __aexit__(self, *tb):
        if self.path is not None:
            async with anyio.open_cancel_scope(shield=True):
                try:
                    await self.stream.aclose()
                except AttributeError:
                    await self.stream.close()


class MsgReader(_MsgRW):
    """Read a stream of messages (encoded with MsgPack) from a file.

    Usage::

        async with MsgReader(path="/tmp/msgs.pack") as f:
            async for msg in f:
                process(msg)

    Arguments:
      buflen (int): The read buffer size. Defaults to 4k.
      path (str): the file to write to.
      stream: the stream to write to.

    Exactly one of ``path`` and ``stream`` must be used.
    """

    _mode = "rb"

    def __init__(self, *a, buflen=4096, **kw):
        super().__init__(*a, **kw)
        self.buflen = buflen

        from .codec import stream_unpacker

        self.unpack = stream_unpacker()

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                msg = next(self.unpack)
            except StopIteration:
                pass
            else:
                return msg

            d = await self.stream.read(self.buflen)
            if d == b"":
                raise StopAsyncIteration
            self.unpack.feed(d)


packer = None


class MsgWriter(_MsgRW):
    """Write a stream of messages to a file (encoded with MsgPack).

    Usage::

        async with MsgWriter("/tmp/msgs.pack") as f:
            for msg in some_source_of_messages():  # or "async for"
                await f(msg)

    Arguments:
      buflen (int): The buffer size. Defaults to 64k.
      path (str): the file to write to.
      stream: the stream to write to.

    Exactly one of ``path`` and ``stream`` must be used.

    The stream is buffered. Call :meth:`distkv.util.MsgWriter.flush` to flush the buffer.
    """

    _mode = "wb"

    def __init__(self, *a, buflen=65536, **kw):
        super().__init__(*a, **kw)

        self.buf = []
        self.buflen = buflen
        self.curlen = 0
        self.excess = 0

        global packer  # pylint: disable=global-statement
        if packer is None:
            from .codec import packer  # pylint: disable=redefined-outer-name

    async def __aexit__(self, *tb):
        async with anyio.fail_after(2, shield=True):
            if self.buf:
                await self.stream.write(b"".join(self.buf))
            await super().__aexit__(*tb)

    async def __call__(self, msg):
        """Write a message (bytes) to the buffer.

        Flushing writes a multiple of ``buflen`` bytes."""
        msg = packer(msg)  # pylint: disable=not-callable
        self.buf.append(msg)
        self.curlen += len(msg)
        if self.curlen + self.excess >= self.buflen:
            buf = b"".join(self.buf)
            pos = self.buflen * int((self.curlen + self.excess) / self.buflen)
            assert pos > 0
            wb, buf = buf[:pos], buf[pos:]
            self.curlen = len(buf)
            self.buf = [buf]
            self.excess = 0
            await self.stream.write(wb)

    async def flush(self):
        """Flush the buffer."""
        if self.buf:
            buf = b"".join(self.buf)
            self.buf = []
            self.excess = (self.excess + len(buf)) % self.buflen
            await self.stream.write(buf)


class _Server:
    _servers = None
    recv_q = None

    def __init__(self, tg, port=0, ssl=None, **kw):
        self.tg = tg
        self.port = port
        self.ports = None
        self._kw = kw
        self.ssl = ssl

    async def _accept(self, server, q):
        self.ports.append(server.socket.getsockname())
        try:
            while True:
                conn = await server.accept()
                if self.ssl:
                    conn = trio.SSLStream(conn, self.ssl, server_side=True)
                await q.send(conn)
        finally:
            async with anyio.fail_after(2, shield=True):
                await q.aclose()
                await server.aclose()

    async def __aenter__(self):
        if current_async_library() != "trio":
            raise RuntimeError("This only works with Trio right now.Sorry.")
        send_q, self.recv_q = trio.open_memory_channel(1)
        try:
            servers = await trio.open_tcp_listeners(self.port, **self._kw)
        except EnvironmentError as exc:
            err = OSError(f"Port {self.port} in use")
            err.errno = exc.errno
            raise err from exc

        self.ports = []
        async with send_q:
            for s in servers:
                await self.tg.spawn(self._accept, s, send_q.clone())
        return self

    async def __aexit__(self, *tb):
        await self.tg.cancel_scope.cancel()
        async with anyio.fail_after(2, shield=True):
            await self.recv_q.aclose()

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self.recv_q.receive()
        except trio.EndOfChannel:
            raise StopAsyncIteration


@asynccontextmanager
async def create_tcp_server(**args) -> _Server:
    async with anyio.create_task_group() as tg:
        server = _Server(tg, **args)
        async with server:
            yield server


def gen_ssl(
    ctx: Union[bool, SSLContext, Dict[str, str]] = False, server: bool = True
) -> Optional[SSLContext]:
    """
    Generate a SSL config from the given context.

    Args:
      ctx: either a Bool (ssl yes/no) or a dict with "key" and "cert" entries.
      server: a flag whether to behave as a server.
    """
    if not ctx:
        return None
    if ctx is True:
        ctx = dict()
    if not isinstance(ctx, dict):
        return ctx

    # pylint: disable=no-member
    ctx_ = trio.ssl.create_default_context(
        purpose=trio.ssl.Purpose.CLIENT_AUTH if server else trio.ssl.Purpose.SERVER_AUTH
    )
    if "key" in ctx:
        ctx_.load_cert_chain(ctx["cert"], ctx["key"])
    return ctx_


def num2byte(num: int, length=None):
    if length is None:
        length = (num.bit_length() + 7) // 8
    return num.to_bytes(length=length, byteorder="big")


def byte2num(data: bytes):
    return int.from_bytes(data, byteorder="big")


def split_one(p, kw):
    """Split 'p' and add to dict 'kw'."""
    try:
        k, v = p.split("=", 1)
    except ValueError:
        if p[-1] == "?":
            k = p[:-1]
            v = getpass(k + "? ")
        else:
            raise
    else:
        if k[-1] == "?":
            k = k[:-1]
            v = getpass(v + ": ")
        try:
            v = int(v)
        except ValueError:
            pass
    kw[k] = v


def _call_proc(code, variables, *a, **kw):
    v = variables[len(a) :]
    if v:
        a = list(a)
        for k in v:
            a.append(kw.pop(k, None))
    eval(code, kw)  # pylint: disable=eval-used
    code = kw["_proc"]
    return code(*a)


def make_proc(code, variables, path, *, use_async=False):  # pylint: disable=redefined-builtin
    """Compile this code block to a procedure.

    Args:
        code: the code block to execute. Text, will be indented.
        vars: variable names to pass into the code
        path: the location where the code is stored
        use_async: False if sync code, True if async, None if in thread
    Returns:
        the procedure to call. All keyval arguments will be in the local
        dict.
    """
    hdr = f"""\
def _proc({ ",".join(variables) }):
    """

    if use_async:
        hdr = "async " + hdr
    code = hdr + code.replace("\n", "\n    ")
    code = compile(code, str(path), "exec")

    return partial(_call_proc, code, variables)


class Module(ModuleType):
    def __repr__(self):
        return "<Module %s>" % (self.__class__.__name__,)


def make_module(code, path):
    """Compile this code block to something module-ish.

    Args:
        code: the code block to execute
        path: the location where the code is / shall be stored
    Returns:
        the procedure to call. All keyval arguments will be in the local
        dict.
    """
    name = ".".join(str(x) for x in path)
    code = compile(code, name, "exec")
    m = sys.modules.get(name, None)
    if m is None:
        m = ModuleType(name)
    eval(code, m.__dict__)  # pylint: disable=eval-used
    sys.modules[name] = m
    return m


class Cache:
    """
    A quick-and-dirty cache that keeps the last N entries of anything
    in memory so that ref and WeakValueDictionary don't lose them.

    Entries get refreshed when they're in the last third of the cache; as
    they're not removed, the actual cache size might only be 2/3rd of SIZE.
    """

    def __init__(self, size):
        self._size = size
        self._head = 0
        self._tail = 0
        self._attr = "_cache__pos"
        self._q = deque()

    def keep(self, entry):
        if getattr(entry, self._attr, -1) > self._tail + self._size / 3:
            return
        self._head += 1
        setattr(entry, self._attr, self._head)
        self._q.append(entry)
        self._flush()

    def _flush(self):
        while self._head - self._tail > self._size:
            self._q.popleft()
            self._tail += 1

    def resize(self, size):
        """Change the size of this cache.
        """
        self._size = size
        self._flush()

    def clear(self):
        while self._head > self._tail:
            self._q.popleft()
            self._tail += 1


@singleton
class NoLock:
    """A dummy singleton that can replace a lock.

    Usage::

        with NoLock if _locked else self._lock:
            pass
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, *tb):
        return


async def data_get(
    obj,
    path,
    *,
    recursive=True,
    as_dict="_",
    maxdepth=-1,
    mindepth=0,
    empty=False,
    raw=False,
    internal=False,
    path_mangle=None,
    item_mangle=None,
):
    """Generic code to dump a subtree.

    `path_mangle` accepts a path and the as_dict parameter. It should
    return the new path. This is used for e.g. prefixing the path with a
    device name. Returning ``None`` causes the entry to be skipped.
    """
    if path_mangle is None:
        path_mangle = lambda x: x
    if item_mangle is None:

        async def item_mangle(x):  # pylint: disable=function-redefined
            return x

    if recursive:
        kw = {}
        if maxdepth is not None:
            kw["max_depth"] = maxdepth
        if mindepth is not None:
            kw["min_depth"] = mindepth
        if empty:
            kw["add_empty"] = True
        if obj.meta:
            kw.setdefault("nchain", obj.meta)
        y = {}
        if internal:
            res = await obj.client._request(action="get_tree_internal", path=path, iter=True, **kw)
        else:
            res = obj.client.get_tree(path, nchain=obj.meta, **kw)
        async for r in res:
            r = await item_mangle(r)
            if r is None:
                continue
            r.pop("seq", None)
            path = r.pop("path")
            path = path_mangle(path)
            if path is None:
                continue

            if as_dict is not None:
                yy = y
                for p in path:
                    yy = yy.setdefault(p, {})
                try:
                    yy[as_dict] = r if obj.meta else r.value
                except AttributeError:
                    if empty:
                        yy[as_dict] = None
            else:
                if raw:
                    y = path
                else:
                    y = {}
                    try:
                        y[path] = r if obj.meta else r.value
                    except AttributeError:
                        if empty:
                            y[path] = None
                        else:
                            continue
                yprint([y], stream=obj.stdout)

        if as_dict is not None:
            if maxdepth:

                def simplex(d):
                    for k, v in d.items():
                        if isinstance(v, dict):
                            d[k] = simplex(d[k])
                    if as_dict in d and d[as_dict] is None:
                        if len(d) == 1:
                            return None
                        else:
                            del d[as_dict]
                    return d

                y = simplex(y)
            yprint(y, stream=obj.stdout)
        return

    if maxdepth is not None or mindepth is not None:
        raise click.UsageError("'mindepth' and 'maxdepth' only work with 'recursive'")
    if as_dict is not None:
        raise click.UsageError("'as-dict' only works with 'recursive'")
    res = await obj.client.get(path, nchain=obj.meta)
    if not obj.meta:
        try:
            res = res.value
        except AttributeError:
            if obj.debug:
                print("No data at", path, file=sys.stderr)
            sys.exit(1)

    if not raw:
        yprint(res, stream=obj.stdout)
    elif isinstance(res, bytes):
        os.write(obj.stdout.fileno(), res)
    else:
        obj.stdout.write(str(res))


@asynccontextmanager
async def as_service(obj=None):
    """
    This async context manager provides readiness and keepalive messages to
    systemd.

    Arguments:
        obj: command context. Needs a ``debug`` attribute.

    The CM yields a (duck-typed) event whose async ``set`` method will
    trigger a ``READY=1`` mesage to systemd.
    """
    from systemd.daemon import notify  # pylint: disable=no-name-in-module

    async def run_keepalive(usec):
        usec /= 1_500_000  # 2/3rd of usec ⇒ sec
        pid = os.getpid()
        while os.getpid() == pid:
            notify("WATCHDOG=1")
            await anyio.sleep(usec)

    def need_keepalive():
        pid = os.getpid()
        epid = int(os.environ.get("WATCHDOG_PID", pid))
        if pid == epid:
            return int(os.environ.get("WATCHDOG_USEC", 0))

    class RunMsg:
        def __init__(self, obj):
            self.obj = obj

        async def set(self):
            notify("READY=1")
            if self.obj is not None and self.obj.debug:
                print("Running.")

    async with anyio.create_task_group() as tg:
        usec = need_keepalive()
        if usec:
            await tg.spawn(run_keepalive, usec)
        try:
            yield RunMsg(obj)
        finally:
            async with anyio.fail_after(2, shield=True):
                await tg.cancel_scope.cancel()


@attr.s
class ValueEvent:
    """A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event`.

    An event object manages an internal value, which is initially
    unset, and a task can wait for it to become True.

    Args:
      ``scope``:  A cancelation scope that will be cancelled if/when
                  this ValueEvent is. Used for clean cancel propagation.

    Note that the value can only be read once.
    """

    event = attr.ib(factory=anyio.create_event, init=False)
    value = attr.ib(default=None, init=False)
    scope = attr.ib(default=None, init=True)

    async def set(self, value):
        """Set the result to return this value, and wake any waiting task.
        """
        self.value = outcome.Value(value)
        await self.event.set()

    async def set_error(self, exc):
        """Set the result to raise this exceptio, and wake any waiting task.
        """
        self.value = outcome.Error(exc)
        await self.event.set()

    def is_set(self):
        """Check whether the event has occurred.
        """
        return self.value is not None

    async def cancel(self):
        """Send a cancelation to the recipient.

        TODO: Trio can't do that cleanly.
        """
        if self.scope is not None:
            await self.scope.cancel()
        await self.set_error(CancelledError())

    async def get(self):
        """Block until the value is set.

        If it's already set, then this method returns immediately.

        The value can only be read once.
        """
        await self.event.wait()
        return self.value.unwrap()


async def spawn(taskgroup, proc, *args, **kw):
    """
    Run a task within this object's task group.

    Returns:
        a cancel scope you can use to stop the task.
    """

    scope = None

    async def _run(proc, args, kw, evt):
        """
        Helper for starting a task.

        This accepts a :class:`ValueEvent`, to pass the task's cancel scope
        back to the caller.
        """
        nonlocal scope
        async with anyio.open_cancel_scope() as sc:
            scope = sc
            await evt.set()
            await proc(*args, **kw)

    evt = anyio.create_event()
    await taskgroup.spawn(_run, proc, args, kw, evt)
    await evt.wait()
    return scope


_PartRE = re.compile("[^:._]+|_|:|\\.")


@total_ordering
class Path(collections.abc.Sequence):
    """
    This object represents the path to a DistKV node.

    It is an immutable list with special representation.
    """

    def __init__(self, *a):
        self._data = a

    @classmethod
    def build(cls, data):
        """Optimized shortcut to generate a path from an existing tuple"""
        if isinstance(data, Path):
            return data
        if not isinstance(data, tuple):
            return cls(*data)
        p = object.__new__(cls)
        p._data = data
        return p

    def __str__(self):
        def _escol(x, spaces=True):  # XXX make the default adjustable?
            x = x.replace(":", "::").replace(".", ":.")
            if spaces:
                x = x.replace(" ", ":_")
            return x

        res = []
        if not self._data:
            return ":"
        for x in self._data:
            if isinstance(x, str) and len(x):
                if res:
                    res.append(".")
                res.append(_escol(x))
            elif isinstance(x, (Path, tuple)) and len(x):
                x = ",".join(repr(y) for y in x)
                res.append(":" + _escol(x))
            elif x is True:
                res.append(":t")
            elif x is False:
                res.append(":f")
            elif x is None:
                res.append(":n")
            elif x == "":
                res.append(":e")
            else:
                if isinstance(x, (Path, tuple)):  # no spaces
                    assert not len(x)
                    x = "()"
                else:
                    x = repr(x)
                res.append(":" + _escol(x))
        return "".join(res)

    def __getitem__(self, x):
        if isinstance(x, slice) and x.start in (0, None) and x.step in (1, None):
            return type(self)(*self._data[x])
        else:
            return self._data[x]

    def __len__(self):
        return len(self._data)

    def __bool__(self):
        return True

    def __eq__(self, other):
        if isinstance(other, Path):
            other = other._data
        return self._data == other

    def __lt__(self, other):
        if isinstance(other, Path):
            other = other._data
        return self._data < other

    def __hash__(self):
        return hash(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __contains__(self, x):
        return x in self._data

    def __or__(self, other):
        return Path(*self._data, other)

    def __add__(self, other):
        if isinstance(other, Path):
            other = other._data
        if not len(other):
            return self
        return Path(*self._data, *other)

    # TODO add alternate output with hex integers

    def __repr__(self):
        return "P(%r)" % (str(self),)

    @classmethod
    def from_str(cls, path):
        """
        Constructor to build a Path from its string representation.
        """
        res = []
        part: Union[NoneType, bool, str] = False
        # non-empty string: accept colon-eval or dot (inline)
        # True: require dot or colon-eval (after :t)
        # False: accept only colon-eval (start)
        # None: accept neither (after dot)

        esc: bool = False
        # marks that an escape char has been seen

        eval_: Union[bool, int] = False
        # marks whether the current input shall be evaluated;
        # 2=it's a hex number

        pos = 0
        if isinstance(path, (tuple, list)):
            return cls.build(path)
        if path == ":":
            return cls()

        def add(x):
            nonlocal part
            if not isinstance(part, str):
                part = ""
            try:
                part += x
            except TypeError:
                raise SyntaxError(f"Cannot add {x!r} at {pos}")

        def done(new_part):
            nonlocal part
            nonlocal eval_
            if isinstance(part, str):
                if eval_:
                    try:
                        if eval_ == 2:
                            part = int(part, 16)
                        else:
                            part = path_eval(part)
                    except Exception:
                        raise SyntaxError(f"Cannot eval {part!r} at {pos}")
                    eval_ = False
                res.append(part)
            part = new_part

        def new(x, new_part):
            nonlocal part
            if part is None:
                raise SyntaxError(f"Cannot use {part!r} at {pos}")
            done(new_part)
            res.append(x)

        if path == "":
            raise SyntaxError("The empty string is not a path")
        for e in _PartRE.findall(path):
            if esc:
                esc = False
                if e in ":.":
                    add(e)
                elif e == "e":
                    new("", True)
                elif e == "t":
                    new(True, True)
                elif e == "f":
                    new(False, True)
                elif e == "n":
                    new(None, True)
                elif e == "_":
                    add(" ")
                elif e[0] == "x":
                    done(None)
                    part = e[1:]
                    eval_ = 2
                else:
                    if part is None:
                        raise SyntaxError(f"Cannot parse {path!r} at {pos}")
                    done("")
                    add(e)
                    eval_ = True
            else:
                if e == ".":
                    if not part:
                        raise SyntaxError(f"Cannot parse {path!r} at {pos}")
                    done(None)
                    pos += 1
                    continue
                elif e == ":":
                    esc = True
                    pos += 1
                    continue
                elif part is True:
                    raise SyntaxError(f"Cannot parse {path!r} at {pos}")
                else:
                    add(e)
            pos += len(e)
        if esc or part is None:
            raise SyntaxError(f"Cannot parse {path!r} at {pos}")
        done(None)
        return cls(*res)

    @classmethod
    def _make(cls, loader, node):
        value = loader.construct_scalar(node)
        return cls.from_str(value)


P = Path.from_str


class PathNode(yaml.nodes.ScalarNode):
    pass


def _path_repr(dumper, data):
    return dumper.represent_scalar("!P", str(data))
    # return ScalarNode(tag, value, style=style)
    # return yaml.events.ScalarEvent(anchor=None, tag='!P', implicit=(True, True), value=str(data))


SafeRepresenter.add_representer(Path, _path_repr)
SafeConstructor.add_constructor("!P", Path._make)


def _bin_from_ascii(loader, node):
    value = loader.construct_scalar(node)
    return value.encode("ascii")


def _bin_to_ascii(dumper, data):
    try:
        data = data.decode("ascii")
    except UnicodeError:
        return dumper.represent_binary(data)
    else:
        return dumper.represent_scalar("!bin", data)


SafeRepresenter.add_representer(bytes, _bin_to_ascii)
SafeConstructor.add_constructor("!bin", _bin_from_ascii)


# path_eval is a simple "eval" replacement to implement resolving
# expressions in paths. While it can be used for math its primary function
_eval = simpleeval.SimpleEval(functions={})
_eval.nodes[_ast.Tuple] = lambda node: tuple(_eval._eval(x) for x in node.elts)
path_eval = _eval.eval


class PathShortener:
    """This class shortens path entries so that the initial components that
    are equal to the last-used path (or the original base) are skipped.

    It is illegal to path-shorten messages whose path does not start with
    the initial prefix.

    Example: The sequence

        a b
        a b c d
        a b c e f
        a b c e g h
        a b c i
        a b j

    is shortened to

        0
        0 c d
        1 e f
        2 g h
        1 i
        0 j

    where the initial number is the passed-in ``depth``, assuming the
    PathShortener is initialized with ``('a','b')``.

    Usage::

        >>> d = _PathShortener(['a','b'])
        >>> d({'path': 'a b c d'.split})
        {'depth':0, 'path':['c','d']}
        >>> d({'path': 'a b c e f'.split})
        {'depth':1, 'path':['e','f']}

    etc.

    Note that the input dict is modified in-place.

    """

    def __init__(self, prefix):
        self.prefix = prefix
        self.depth = len(prefix)
        self.path = []

    def __call__(self, res):
        try:
            p = res["path"]
        except KeyError:
            return
        if list(p[: self.depth]) != list(self.prefix):
            raise RuntimeError(f"Wrong prefix: has {p!r}, want {self.prefix!r}")

        p = p[self.depth :]  # noqa: E203
        cdepth = min(len(p), len(self.path))
        for i in range(cdepth):
            if p[i] != self.path[i]:
                cdepth = i
                break
        self.path = p
        p = p[cdepth:]
        res["path"] = p
        res["depth"] = cdepth


class PathLongener:
    """
    This reverts the operation of a PathShortener. You need to pass the
    same prefix in.

    Calling a PathLongener with a dict without ``depth`` or ``path``
    attributes is a no-op.
    """

    def __init__(self, prefix: Union[Path, tuple] = ()):
        self.depth = len(prefix)
        self.path = Path.build(prefix)

    def __call__(self, res):
        p = res.get("path", None)
        if p is None:
            return
        d = res.pop("depth", None)
        if d is None:
            return
        if not isinstance(p, tuple):
            # may be a list, dammit
            p = tuple(p)
        p = self.path[: self.depth + d] + p
        self.path = p
        res["path"] = p


def res_get(res, attr: Path, **kw):  # pylint: disable=redefined-outer-name
    """
    Get a node's value and access the dict items beneath it.

    The node value must be an attrdict.
    """
    val = res.get("value", None)
    if val is None:
        return None
    return val._get(attr, **kw)


def res_update(res, attr: Path, value=None, **kw):  # pylint: disable=redefined-outer-name
    """
    Set a node's sub-item's value, possibly merging dicts.
    Entries set to 'NotGiven' are deleted.

    The node value must be an attrdict.

    Returns the new value.
    """
    val = res.get("value", attrdict())
    return val._update(attr, value=value, **kw)


def res_delete(res, attr: Path, **kw):  # pylint: disable=redefined-outer-name
    """
    Remove a node's sub-item's value, possibly removing now-empty
    intermediate dicts.

    The node value must be an attrdict.

    Returns the new value.
    """
    val = res.get("value", attrdict())
    return val._delete(attr, **kw)


def logger_for(path: Path):
    """
    Create a logger for this ``path``.
    """
    if not len(path):
        p = "distkv.root"
    elif path[0] is None:
        p = "distkv.meta"
    elif path[0] == ".distkv":
        p = "distkv.sub"
    else:
        p = "distkv.at"
    if len(path) > 1:
        p += "." + ".".join(path[1:])
    return logging.getLogger(p)


def digits(n, digits=6):  # pylint: disable=redefined-outer-name
    """
    Returns ``n`` rounded to ``digits`` significant digits. Default: 6.
    Ensures that the number doesn't carry nonsense precision or
    floating-point artefacts.

    >>> digits(123456789, 4)
    123400000
    >>> digits(math.pi, 4)
    3.142

    ``digits`` may be a fraction, in order to move the cut-off point to
    somewhere other than between 9.999 and 10.00.
    """
    return round(n, int(digits - 1 - log10(abs(n))))



# command line interface

import time
import asyncclick as click
import datetime

from distkv.util import PathLongener, MsgReader, NotGiven, yprint, data_get, P
from distkv.client import StreamedRequest
from distkv.command import node_attr

import logging

logger = logging.getLogger(__name__)


@main.group(short_help="Manage data.")  # pylint: disable=undefined-variable
async def cli():
    """
    This subcommand accesses the actual user data stored in your DistKV tree.
    """
    pass


@cli.command()
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m", "--maxdepth", type=int, default=None, help="Limit recursion depth. Default: whole tree"
)
@click.option(
    "-M", "--mindepth", type=int, default=None, help="Starting depth. Default: whole tree"
)
@click.option("-r", "--recursive", is_flag=True, help="Read a complete subtree")
@click.option("-e", "--empty", is_flag=True, help="Include empty nodes")
@click.option("-R", "--raw", is_flag=True, help="Print string values without quotes etc.")
@click.argument("path", nargs=1)
@click.pass_obj
async def get(obj, path, **k):
    """
    Read a DistKV value.

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    path = P(path)
    await data_get(obj, path, **k)


@cli.command("list")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=1,
    help="Limit recursion depth. Default: 1 (single layer).",
)
@click.option(
    "-M", "--mindepth", type=int, default=1, help="Starting depth. Default: 1 (single layer)."
)
@click.argument("path", nargs=1)
@click.pass_obj
async def list_(obj, path, **k):
    """
    List DistKV values.

    This is like "get" but with "--mindepth=1 --maxdepth=1 --recursive --empty"

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    path = P(path)
    k["recursive"] = True
    k["raw"] = True
    k["empty"] = True
    await data_get(obj, path, **k)


@cli.command("set", short_help="Add or update an entry")
@click.option("-v", "--value", help="The value to store. Mandatory.")
@click.option("-e", "--eval", "eval_", is_flag=True, help="The value shall be evaluated.")
@click.option("-P", "--path", "path_", is_flag=True, help="The value is a path.")
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-n", "--new", is_flag=True, help="This is a new entry.")
@click.option("-a", "--attr", default=None, help="Modify attribute.")
@click.argument("path", nargs=1)
@click.pass_obj
async def set_(obj, path, value, eval_, last, new, path_, attr):
    """
    Store a value at some DistKV position.

    If you update a value, you should use "--last"
    to ensure that no other change arrived.

    When adding a new entry, use "--new" to ensure that you don't
    accidentally overwrite something.
    """
    path = P(path)
    if attr is not None:
        attr = P(attr)
    if eval_:
        if path_:
            raise click.UsageError("'eval' and 'path' are mutually exclusive")
        if value == "-":
            value = NotGiven
        else:
            value = eval(value)  # pylint: disable=eval-used
            eval_ = False
    elif path_:
        value = P(value)
    args = {}
    if new:
        if last:
            raise click.UsageError("'new' and 'last' are mutually exclusive")
        args["chain"] = None
    else:
        if last:
            args["chain"] = {"node": last[0], "tick": int(last[1])}

    if attr is not None:
        res = await node_attr(obj, path, attr, value, eval_=eval_, **args)
    else:
        res = await obj.client.set(path, value=value, nchain=obj.meta, **args)

    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command(short_help="Delete an entry / subtree")
@click.argument("path", nargs=1)
@click.option("-p", "--prev", default=NotGiven, help="Previous value. Deprecated; use 'last'")
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-r", "--recursive", is_flag=True, help="Delete a complete subtree")
@click.option("--internal", is_flag=True, help="Affect the internal tree. DANGER.")
@click.option("-e", "--eval", "eval_", is_flag=True, help="The previous value shall be evaluated.")
@click.pass_obj
async def delete(obj, path, prev, last, recursive, eval_, internal):
    """
    Delete an entry, or a whole subtree.

    You really should use "--last" (preferred) or "--prev" (if you must) to
    ensure that no other change arrived (but note that this doesn't work
    when deleting a subtree).

    Non-recursively deleting an entry with children works and does *not*
    affect the child entries.

    The root entry cannot be deleted.
    """
    path = P(path)
    args = {}
    if eval_ and prev is NotGiven:
        raise click.UsageError("You need to add a value that can be evaluated")
    if recursive:
        if prev is not NotGiven or last:
            raise click.UsageError("'local' and 'force' are mutually exclusive")
        if internal:
            raise click.UsageError("'internal' and 'recursive' are mutually exclusive")
    else:
        if prev is not NotGiven:
            if eval_:
                prev = eval(prev)  # pylint: disable=eval-used
            args["prev"] = prev
        if last:
            args["chain"] = {"node": last[0], "tick": int(last[1])}

    res = await obj.client._request(
        action="delete_tree" if recursive else "delete_internal" if internal else "delete_value",
        path=path,
        nchain=obj.meta,
        **args
    )
    if isinstance(res, StreamedRequest):
        pl = PathLongener(path)
        async for r in res:
            pl(r)
            if obj.meta:
                yprint(r, stream=obj.stdout)
    else:
        if obj.meta:
            yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-s", "--state", is_flag=True, help="Also get the current state.")
@click.option("-o", "--only", is_flag=True, help="Value only, nothing fancy.")
@click.argument("path", nargs=1)
@click.pass_obj
async def watch(obj, path, state, only):
    """Watch a DistKV subtree"""

    flushing = not state
    path = P(path)
    seen = False

    async with obj.client.watch(
        path, nchain=obj.meta, fetch=state, max_depth=0 if only else -1
    ) as res:
        async for r in res:
            if r.get("state", "") == "uptodate":
                if only and not seen:
                    # value doesn't exist
                    return
                flushing = True
            else:
                del r["seq"]
                if only:
                    try:
                        print(r.value, file=obj.stdout)
                        continue
                    except AttributeError:
                        # value has been deleted
                        return
            if flushing:
                r["time"] = time.monotonic()
                r["date"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            yprint(r, stream=obj.stdout)
            print("---", file=obj.stdout)
            if flushing:
                obj.stdout.flush()
            seen = True


@cli.command()
@click.option("-i", "--infile", type=click.Path(), help="File to read (msgpack).")
@click.argument("path", nargs=1)
@click.pass_obj
async def update(obj, path, infile):
    """Send a list of updates to a DistKV subtree"""
    path = P(path)
    async with MsgReader(path=infile) as reader:
        async for msg in reader:
            await obj.client.set(*path, *msg.path, value=msg.value)



import os
import logging

logger = logging.getLogger(__name__)

from distkv.util import yload


def load_cfg(cfg):  # pylint: disable=redefined-outer-name
    if os.path.exists(cfg):
        pass
    elif os.path.exists(os.path.join("tests", cfg)):  # pragma: no cover
        cfg = os.path.join("tests", cfg)
    elif os.path.exists(os.path.join(os.pardir, cfg)):  # pragma: no cover
        cfg = os.path.join(os.pardir, cfg)
    else:  # pragma: no cover
        raise RuntimeError(f"Config file {cfg!r} not found")

    with open(cfg) as f:
        cfg = yload(f)

    from logging.config import dictConfig

    cfg["disable_existing_loggers"] = False
    try:
        dictConfig(cfg)
    except ValueError:
        pass
    logging.captureWarnings(True)
    logger.debug("Test %s", "starting up")
    return cfg


cfg = load_cfg(os.environ.get("LOG_CFG", "logging.cfg"))


import trio._core._run as tcr

if "PYTHONHASHSEED" in os.environ:
    tcr._ALLOW_DETERMINISTIC_SCHEDULING = True
    tcr._r.seed(os.environ["PYTHONHASHSEED"])



from pytest_trio.enable_trio_mode import *  # noqa: E501,F403,F401 # pylint:disable=wildcard-import,unused-wildcard-import
import pytest  # noqa: F403,F401  pylint:disable=W0611

import logging  # noqa: F403,F401  pylint:disable=W0611

# logging.basicConfig(level=logging.DEBUG)



# command line interface

import asyncclick as click

from range_set import RangeSet
from distkv.util import yprint, PathLongener, P
from collections.abc import Mapping

import logging

logger = logging.getLogger(__name__)


@main.group(short_help="Control internal state.")  # pylint: disable=undefined-variable
async def cli():
    """
    This subcommand queries and controls the server's internal state.
    """
    pass


@cli.command()
@click.option("-n", "--nodes", is_flag=True, help="Get node status.")
@click.option("-d", "--deleted", is_flag=True, help="Get deletion status.")
@click.option("-m", "--missing", is_flag=True, help="Get missing-node status.")
@click.option(
    "-r",
    "--remote-missing",
    "remote_missing",
    is_flag=True,
    help="Get remote-missing-node status.",
)
@click.option("-p", "--present", is_flag=True, help="Get known-data status.")
@click.option("-s", "--superseded", is_flag=True, help="Get superseded-data status.")
@click.option("-D", "--debug", is_flag=True, help="Get internal verbosity.")
@click.option("--debugger", is_flag=True, help="Start a remote debugger. DO NOT USE.")
@click.option("-k", "--known", hidden=True, is_flag=True, help="Get superseded-data status.")
@click.option("-a", "--all", is_flag=True, help="All available data.")
@click.pass_obj
async def state(obj, **flags):
    """
    Dump the server's state.
    """
    if flags.pop("known", None):
        flags["superseded"] = True
    if flags.pop("all", None):
        flags["superseded"] = True
        flags["present"] = True
        flags["nodes"] = True
        flags["deleted"] = True
        flags["missing"] = True
        flags["remote_missing"] = True
    res = await obj.client._request("get_state", iter=False, **flags)
    k = res.pop("known", None)
    if k is not None:
        res["superseded"] = k
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-d", "--deleted", is_flag=True, help="Mark as deleted. Default: superseded")
@click.option(
    "-n", "--node", "source", default="?", help="The node this message is faked as being from."
)
@click.option("-b", "--broadcast", is_flag=True, help="Send to all servers")
@click.argument("node", nargs=1)
@click.argument("items", type=int, nargs=-1)
@click.pass_obj
async def mark(obj, deleted, source, node, items, broadcast):
    """
    Fix internal state. Use no items to fetch the current list from the
    server's ``missing`` state. Use an empty node name to add the whole
    list, not just a single node's.

    This is a dangerous command.
    """

    k = "deleted" if deleted else "superseded"
    if not items:
        r = await obj.client._request("get_state", iter=False, missing=True)
        r = r["missing"]
        if node != "":
            r = {node: r[node]}
    elif node == "":
        raise click.UsageError("You can't do that with an empty node")
    else:
        r = RangeSet()
        for i in items:
            r.add(i)
        r = {node: r.__getstate__()}

    msg = {k: r, "node": source}

    await obj.client._request("fake_info", iter=False, **msg)
    if broadcast:
        await obj.client._request("fake_info_send", iter=False, **msg)

    res = await obj.client._request("get_state", iter=False, **{k: True})
    yprint(res, stream=obj.stdout)


@cli.command(short_help="Manage the Deleter list")
@click.option("-d", "--delete", is_flag=True, help="Remove these nodes")
@click.argument("nodes", nargs=-1)
@click.pass_obj
async def deleter(obj, delete, nodes):
    """
    Manage the Deleter list

    This is the set of nodes that must be online for removal of deleted
    entries from DistKV's data.

    There should be one such node in every possible network partition.
    Also, all nodes with permanent storage should be on the list.

    Usage:
    - … deleter          -- list state
    - … deleter NODE…    -- add this node
    - … deleter -d NODE… -- remove this node
    """

    res = await obj.client._request(
        action="get_internal",
        path=("actor", "del"),
        iter=False,
        nchain=3 if delete or nodes else 2,
    )
    val = res.get("value", {})
    if isinstance(val, Mapping):
        val = val.get("nodes", [])
    # else: compatibility, TODO remove
    val = set(val)
    if delete:
        if nodes:
            val -= set(nodes)
        else:
            val = set()
    elif nodes:
        val |= set(nodes)
    else:
        yprint(res, stream=obj.stdout)
        return

    val = {"nodes": list(val)}
    res = await obj.client._request(
        action="set_internal", path=("actor", "del"), iter=False, chain=res.chain, value=val
    )
    res.value = val
    yprint(res, stream=obj.stdout)


@cli.command()
@click.argument("path", nargs=1)
@click.pass_obj
async def dump(obj, path):
    """
    Dump internal state.

    This displays DistKV's internal state.
    """

    path = P(path)
    y = {}
    pl = PathLongener()
    async for r in await obj.client._request("get_tree_internal", path=path, iter=True, nchain=0):
        pl(r)
        path = r["path"]
        yy = y
        for p in path:
            yy = yy.setdefault(p, {})
        try:
            yy["_"] = r["value"]
        except KeyError:
            pass
    yprint(y, stream=obj.stdout)


@cli.command()
@click.argument("node", nargs=1)
@click.argument("tick", type=int, nargs=1)
@click.pass_obj
async def get(obj, node, tick):
    """
    Fetch data by node+tick.

    This looks up internal data.
    """

    res = await obj.client._request("get_value", node=node, tick=tick, nchain=99)
    if not obj.meta:
        res = res.value
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-n", "--num", type=int, help="Return at most this many IDs")
@click.option("-c", "--current", is_flag=True, help="Return only IDs with current data")
@click.option("-C", "--copy", is_flag=True, help="Create an no-op change")
@click.argument("node", nargs=1)
@click.pass_obj
async def enum(obj, node, num, current, copy):
    """
    List IDs of live data by a specific node.

    Can be used to determine whether a node still has live data,
    otherwise it can be deleted.

    If '--current' is set, only the IDs of the entries that have last been
    updated by that node are shown. '--copy' rewrites these entries.

    Increase verbosity to also show the oject paths.
    """

    res = await obj.client._request("enum_node", node=node, max=num, current=current)
    if obj.meta and not copy and obj.debug <= 1:
        yprint(res, stream=obj.stdout)
    else:
        for k in res.result:
            if copy or obj.debug > 1:
                res = await obj.client._request("get_value", node=node, tick=k, nchain=3)
                if obj.debug > 1:
                    print(k, res.path)
                else:
                    print(k)
                if copy and res.chain.node == node:
                    res = await obj.client.set(res.path, value=res.value, chain=res.chain)
            else:
                print(k)


@cli.command()
@click.argument("node", nargs=1)
@click.pass_obj
async def kill(obj, node):
    """
    Remove a node from the node list.

    This command only works if this node does not have any current data in
    the system.
    """
    res = await obj.client._request("kill_node", node=node)
    if obj.meta:
        yprint(res, stream=obj.stdout)



import pytest
import trio

from distkv.mock.mqtt import stdtest

from distkv.util import PathLongener, P

from distkv.errors import ErrorRoot
import logging

logger = logging.getLogger(__name__)


async def collect(i, path=()):
    res = []
    pl = PathLongener(path)
    async for r in i:
        r.pop("tock", 0)
        r.pop("seq", 0)
        pl(r)
        res.append(r)
    return res


@pytest.mark.trio
async def test_81_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=40) as st:
        async with st.client() as c:
            e = await ErrorRoot.as_handler(c)
            async with st.client() as cx:
                ex = await ErrorRoot.as_handler(cx)
                try:
                    1 / 0
                except Exception as exc:
                    await ex.record_error("tester", P("here.or.there"), exc=exc)
                await trio.sleep(1)
                n = 0
            for err in e.all_errors("tester"):
                n += 1
                await err.resolve()
            assert n == 1
            await trio.sleep(1)


@pytest.mark.trio
@pytest.mark.xfail
async def test_82_many(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=80) as st:
        async with st.client() as cx, st.client() as cy, st.client() as cz:
            ex = await ErrorRoot.as_handler(cx, name="a1")
            ey = await ErrorRoot.as_handler(cy, name="a2")
            ez = await ErrorRoot.as_handler(cz, name="a3")

            async def err(e):
                with trio.CancelScope(shield=True):
                    await e.record_error(
                        "tester", P("dup"), message="Owchie at {node}", data={"node": e.name}
                    )

            async with trio.open_nursery() as tg:
                tg.start_soon(err, ex)
                tg.start_soon(err, ey)
                tg.start_soon(err, ez)
                await trio.sleep(2)

            await st.run("data get -rd_", do_stdout=False)
            await trio.sleep(2)

            n = 0
            for err in ex.all_errors("tester"):
                n += 1
                logger.warning("DEL ASSERT %d", n)
                assert len(list(err)) == 3, list(err)
                for k in err:
                    assert k._name in {"a1", "a2", "a3"}, k
                await err.resolve()
            assert n == 1
            await trio.sleep(1)



try:
    from contextlib import asynccontextmanager, AsyncExitStack
except ImportError:
    from async_generator import asynccontextmanager
    from async_exit_stack import AsyncExitStack
import os
import trio
import anyio
import mock
import copy
import time
from functools import partial
from asyncscope import main_scope

from distkv.default import CFG
from distkv.server import Server
from distkv.util import attrdict, combine_dict, NotGiven
from distkv.mock import S
from distmqtt.broker import create_broker

import logging

logger = logging.getLogger(__name__)

otm = time.time

PORT = 40000 + (os.getpid() + 10) % 10000

broker_cfg = {
    "listeners": {"default": {"type": "tcp", "bind": f"127.0.0.1:{PORT}"}},
    "timeout-disconnect-delay": 2,
    "auth": {"allow-anonymous": True, "password-file": None},
}

URI = f"mqtt://127.0.0.1:{PORT}/"


@asynccontextmanager
async def stdtest(n=1, run=True, ssl=False, tocks=20, **kw):
    C_OUT = CFG.get("_stdout", NotGiven)
    if C_OUT is not NotGiven:
        del CFG["_stdout"]
    TESTCFG = copy.deepcopy(CFG)
    TESTCFG.server.port = None
    TESTCFG.root = "test"
    if C_OUT is not NotGiven:
        CFG["_stdout"] = C_OUT
        TESTCFG["_stdout"] = C_OUT

    if ssl:
        import ssl
        import trustme

        ca = trustme.CA()
        cert = ca.issue_server_cert("127.0.0.1")
        server_ctx = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        client_ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        ca.configure_trust(client_ctx)
        cert.configure_cert(server_ctx)
    else:
        server_ctx = client_ctx = False

    clock = trio.lowlevel.current_clock()
    # clock.autojump_threshold = 0.1
    # clock.rate = 5

    async def mock_get_host_port(st, host):
        i = int(host[host.rindex("_") + 1 :])  # noqa: E203
        s = st.s[i]
        await s.is_serving
        for host, port, *_ in s.ports:
            if host[0] != ":":
                return host, port

    def tm():
        try:
            return trio.current_time()
        except RuntimeError:
            return otm()

    async def mock_set_tock(self, old):
        assert self._tock < tocks, "Test didn't terminate. Limit:" + str(tocks)
        await old()

    async with main_scope("_distkv_test_mqtt") as scp:
        tg = scp._tg
        st = S(tg, client_ctx)
        async with AsyncExitStack() as ex:
            st.ex = ex  # pylint: disable=attribute-defined-outside-init
            ex.enter_context(mock.patch("time.time", new=tm))
            ex.enter_context(mock.patch("time.monotonic", new=tm))
            logging._startTime = tm()
            await ex.enter_async_context(create_broker(config=broker_cfg))

            args_def = kw.get("args", attrdict())
            for i in range(n):
                name = "test_" + str(i)
                args = kw.get(name, args_def)
                args = combine_dict(
                    args,
                    args_def,
                    {
                        "cfg": {
                            "connect": {"ssl": client_ctx},
                            "server": {
                                "bind_default": {
                                    "host": "127.0.0.1",
                                    "port": i + PORT + 1,
                                    "ssl": server_ctx,
                                },
                                "backend": "mqtt",
                                "mqtt": {"uri": URI},
                            },
                        }
                    },
                    {"cfg": TESTCFG},
                )
                args_def.pop("init", None)
                s = Server(name, **args)
                ex.enter_context(
                    mock.patch.object(s, "_set_tock", new=partial(mock_set_tock, s, s._set_tock))
                )
                ex.enter_context(
                    mock.patch.object(s, "_get_host_port", new=partial(mock_get_host_port, st))
                )
                st.s.append(s)

            evts = []
            for i in range(n):
                if kw.get("run_" + str(i), run):
                    evt = anyio.create_event()
                    await tg.spawn(partial(st.s[i].serve, ready_evt=evt))
                    evts.append(evt)
            for e in evts:
                await e.wait()
            try:
                yield st
            finally:
                async with anyio.fail_after(2, shield=True):
                    logger.info("Runtime: %s", clock.current_time())
                    await tg.cancel_scope.cancel()
        logger.info("End")
        pass  # unwinding AsyncExitStack



# command line interface

import sys
import datetime
import asyncclick as click

from distkv.util import MsgReader, MsgWriter
from distkv.util import yprint, PathLongener, P, yload, Path
from distkv.codec import unpacker

import logging

logger = logging.getLogger(__name__)


@main.group(short_help="Manage data.")  # pylint: disable=undefined-variable
async def cli():
    """
    Low-level tools that don't depend on a running server.
    """
    pass


@cli.command("cfg")
@click.argument("path", nargs=1)
@click.pass_obj
async def cfg_dump(obj, path):
    """Emit the current configuration as a YAML file.

    You can limit the output by path elements.
    E.g., "cfg connect.host" will print "localhost".

    Single values are printed with a trailing line feed.
    """
    cfg = obj.cfg
    for p in P(path):
        try:
            cfg = cfg[p]
        except KeyError:
            if obj.debug:
                print("Unknown:", p)
            sys.exit(1)
    if isinstance(cfg, str):
        print(cfg, file=obj.stdout)
    else:
        yprint(cfg, stream=obj.stdout)


@cli.command("file")
@click.option("-p", "--path", is_flag=True, default=False, help="Unwrap paths")
@click.option("-f", "--filter", "filter_", multiple=True, help="Only emit entries with this path")
@click.argument("file", nargs=1)
@click.pass_obj
async def file_(obj, file, path, filter_):
    """Read a MsgPack file and dump as YAML."""
    if path or filter_:
        pl = PathLongener()
    else:
        pl = lambda _: None
    filter_ = [P(x) for x in filter_]
    async with MsgReader(path=file) as f:
        async for msg in f:
            pl(msg)
            if filter_:
                if "path" not in msg:
                    continue
                for f in filter_:
                    if msg.path[: len(f)] == f:
                        break
                else:
                    continue
            yprint(msg, stream=obj.stdout)
            print("---", file=obj.stdout)


@cli.command("yaml")
@click.argument("msgpack", nargs=1)
async def yaml_(msgpack):
    """Read a YAML file from stdin and dump as msgpack."""
    async with MsgWriter(path=msgpack) as f:
        for d in yload(sys.stdin, multi=True):
            await f(d)


@cli.command()
@click.argument("node", nargs=1)
@click.argument("file", type=click.Path(), nargs=1)
async def init(node, file):
    """Write an initial preload file.

    Usage: distkv dump init <node> <outfile>

    Writes an initial DistKV file that behaves as if it was generated by <node>.

    Using this command, followed by "distkv server -l <outfile> <node>", is
    equivalent to running "distkv server -i 'Initial data' <node>.
    """
    async with MsgWriter(path=file) as f:
        await f(
            dict(
                chain=dict(node=node, tick=1, prev=None),
                depth=0,
                path=[],
                tock=1,
                value="Initial data",
            )
        )


@cli.command("msg")
@click.argument("path", nargs=-1)
@click.pass_obj
async def msg_(obj, path):
    """
    Monitor the server-to-sever message stream.

    The default is the main server's "update" stream.
    Use '+NAME' to monitor a different stream instead.
    Use '+' to monitor all streams.

    Common streams:
    * ping: sync: all servers (default)
    * update: data changes
    * del: sync: nodes responsible for cleaning up deleted records
    """
    from distkv.backend import get_backend

    class _Unpack:
        def __init__(self):
            self._part_cache = dict()

    import distkv.server

    _Unpack._unpack_multiple = distkv.server.Server._unpack_multiple
    _unpacker = _Unpack()._unpack_multiple

    if not path:
        path = P(obj.cfg.server.root) | "update"
        path.append("update")
    elif len(path) == 1:
        path = P(path[0])
        if len(path) == 1 and path[0].startswith("+"):
            p = path[0][1:]
            path = P(obj.cfg.server.root)
            path |= [p or "#"]
    be = obj.cfg.server.backend
    kw = obj.cfg.server[be]

    async with get_backend(be)(**kw) as conn:
        async with conn.monitor(*path) as stream:
            async for msg in stream:
                v = vars(msg)
                if isinstance(v.get("payload"), (bytearray, bytes)):
                    t = msg.topic
                    v = unpacker(v["payload"])
                    v = _unpacker(v)
                    if v is None:
                        continue
                    v["_topic"] = Path.build(t)
                else:
                    v["_type"] = type(msg).__name__
                v["_timestamp"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

                yprint(v, stream=obj.stdout)
                print("---", file=obj.stdout)



import pytest
import trio
import mock

# doesn't work with MQTT because we can't split
from distkv.mock.serf import stdtest
import asyncserf
import msgpack
from distkv.util import attrdict, P, Path

import logging

logger = logging.getLogger(__name__)

N = 10


def _skip_check(self):
    if getattr(self._client, "prefix", "") != "test.core":
        return False
    if not hasattr(self, "_test_foo") or not self._test_foo:
        self._test_foo = ["test_2", "test_3", "test_4"]
    self.logger.debug("SKIP? %s %r %r", self._name, self._test_foo, self._history)
    p = self._test_foo[-1]
    if p in self._history:
        self._test_foo.pop()
        return False
    if self._name == p:
        return True
    return False


@pytest.mark.trio
async def test_10_many(autojump_clock):  # pylint: disable=unused-argument
    """
    This test starts multiple servers at the same time and checks that
    final object deletion works.
    """
    async with stdtest(test_1={"init": 420}, n=N, tocks=1500) as st:
        st.ex.enter_context(mock.patch("asyncactor.Actor._skip_check", new=_skip_check))

        s = st.s[1]
        async with st.client(1) as ci:
            assert (await ci.get(P(":"))).value == 420
            await ci.set(P("ping"), value="pong")
            await ci.set(P("delete.me"), value="later")

            await ci._request(
                "set_internal",
                path=P("actor.del"),
                value={"nodes": "test_2 test_3 test_4".split()},
            )

        await trio.sleep(1)

        async def s1(i, *, task_status=trio.TASK_STATUS_IGNORED):
            async with st.client(i) as c:
                task_status.started()
                assert (await c.get(P(":"))).value == 420
                assert (await c.get(P("ping"))).value == "pong"
                await c.set(Path("foo", i), value=420 + i)

        async with trio.open_nursery() as tg:
            for i in range(1, N):
                await tg.start(s1, i)

        async with st.client(2) as ci:
            assert (await ci.get(P("delete.me"))).value == "later"
            await ci.delete(P("delete.me"))

        await trio.sleep(1)
        NN = min(N - 1, 3)
        for j in [0] + s._actor._rand.sample(range(1, N), NN):
            async with st.client(j) as c:
                for i in s._actor._rand.sample(range(1, N), NN):
                    assert (await c.get(Path("foo", i))).value == 420 + i

        async with st.client(N - 1) as ci:
            r = await ci.get(P("delete.me"), nchain=2)
            assert "value" not in r
            assert r.chain is not None

            with trio.fail_after(9999):
                while True:
                    r = await ci.get(P("delete.me"), nchain=2)
                    if "value" not in r and r.chain is None:
                        break
                    await trio.sleep(10)

        pass  # server end


@pytest.mark.trio
@pytest.mark.parametrize("tocky", [-10, -2, -1, 0, 1, 2, 10])
async def test_11_split1(autojump_clock, tocky):  # pylint: disable=unused-argument
    """
    This test checks that re-joining a split network updates everything.
    """
    n_two = 0

    async with stdtest(test_1={"init": 420}, n=N, tocks=1000) as st:

        async def watch(*, task_status=trio.TASK_STATUS_IGNORED):
            nonlocal n_two
            async with asyncserf.serf_client() as s:
                async with s.stream("user:test.update") as sr:
                    task_status.started()
                    async for r in sr:
                        msg = msgpack.unpackb(
                            r.data, object_pairs_hook=attrdict, raw=False, use_list=False
                        )
                        if msg.get("value", "") == "two":
                            n_two += 1

                    i.s()

        await st.tg.spawn(watch)
        async with st.client(1) as ci:
            assert (await ci.get(P(":"))).value == 420
            r = await ci.set(P("ping"), value="pong")
            pongtock = r.tock
            await ci.set(P("drop.me"), value="here")

        async def s1(i, *, task_status=trio.TASK_STATUS_IGNORED):
            async with st.client(i) as c:
                task_status.started()
                assert (await c.get(P(":"))).value == 420
                await trio.sleep(5)
                r = await c.get(P("ping"))
                assert r.value == "pong"
                assert r.tock == pongtock
                await c.set(Path("foo", i), value=420 + i)
                pass  # client end

        async with trio.open_nursery() as tg:
            for i in range(1, N):
                await tg.start(s1, i)

        await trio.sleep(30)
        st.split(N // 2)
        if tocky:
            async with st.client(2 if tocky < 0 else N - 2) as ci:
                for i in range(abs(tocky)):
                    await ci.set(Path("one", i), value="two")
        await trio.sleep(30)
        async with st.client(1) as ci:
            await ci.delete(P("drop.me"))

        async with st.client(N - 1) as c:
            r = await c.set(P("ping"), value="pongpang")
            pangtock = r.tock

        await trio.sleep(1)
        async with st.client(0) as c:
            # assert that this value is gone
            r = await c.get(P("drop.me"), nchain=3)
            # assert r.chain is None -- not yet
            assert "value" not in r
        await trio.sleep(1)
        async with st.client(N - 1) as c:
            # assert that this value is still here
            r = await c.get(P("drop.me"), nchain=3)
            assert r.chain is not None
            assert r.value == "here"
        await trio.sleep(1)
        async with st.client(0) as c:
            r = await c.get(P("ping"))
            assert r.value == "pong"
            assert r.tock == pongtock

        st.join(N // 2)
        await trio.sleep(200)
        async with st.client(0) as c:
            r = await c.get(P("ping"))
            assert r.value == "pongpang"
            assert r.tock == pangtock

        async with st.client(N - 1) as c:
            # assert that this value is now gone
            r = await c.get(P("drop.me"), nchain=3)
            # assert r.chain is None -- not yet
            assert "value" not in r
        pass  # server end

    # Now make sure that updates are transmitted once
    assert n_two <= 2 * N + 1



# command line interface

import asyncclick as click
import sys

from distkv.util import yprint, NotGiven, yload, P

import logging

logger = logging.getLogger(__name__)


@main.group()  # pylint: disable=undefined-variable
async def cli():
    """Manage code stored in DistKV."""
    pass


@cli.command()
@click.option("-s", "--script", type=click.File(mode="w", lazy=True), help="Save the code here")
@click.argument("path", nargs=1)
@click.pass_obj
async def get(obj, path, script):
    """Read a code entry"""
    path = P(path)
    if not len(path):
        raise click.UsageError("You need a non-empty path.")
    res = await obj.client._request(
        action="get_value", path=obj.cfg["codes"]["prefix"] + path, iter=False, nchain=obj.meta
    )
    if "value" not in res:
        if obj.debug:
            print("No entry here.", file=sys.stderr)
        sys.exit(1)
    if not obj.meta:
        res = res.value
    if script:
        code = res.pop("code", None)
        if code is not None:
            print(code, file=script)
    yprint(res, stream=obj.stdout)


@cli.command("set")
@click.option("-a", "--async", "async_", is_flag=True, help="The code is async")
@click.option("-t", "--thread", is_flag=True, help="The code should run in a worker thread")
@click.option("-s", "--script", type=click.File(mode="r"), help="File with the code")
@click.option("-i", "--info", type=str, help="one-liner info about the code")
@click.option("-d", "--data", type=click.File(mode="r"), help="load the metadata (YAML)")
@click.option("-v", "--var", "vars_", nargs=2, multiple=True, help="Value (name val…)")
@click.option("-e", "--eval", "eval_", nargs=2, multiple=True, help="Value (name val), evaluated")
@click.option("-p", "--path", "path_", nargs=2, multiple=True, help="Value (name val), as path")
@click.argument("path", nargs=1)
@click.pass_obj
async def set_(obj, path, thread, script, data, vars_, eval_, path_, async_, info):
    """Save Python code.

    The code may have inputs. You specify the inputs and their default
    values with '-v VAR VALUE' (string), '-p VAR PATH' (DistKV path), or
    '-e VAR EXPR' (Python expression). Use '-e VAR -' to state that VAR
    shall not have a default value, and '-e VAR /' to delete VAR from the
    list of inputs entirely.
    """
    if async_:
        if thread:
            raise click.UsageError("You can't specify both '--async' and '--thread'.")
    else:
        if thread:
            async_ = False
        else:
            async_ = None

    path = P(path)
    if not len(path):
        raise click.UsageError("You need a non-empty path.")

    if data:
        msg = yload(data)
    else:
        msg = await obj.client.get(obj.cfg["codes"]["prefix"] + path, nchain=3)
    chain = NotGiven
    if "value" in msg:
        chain = msg.get("chain", NotGiven)
        msg = msg["value"]
    if async_ is not None or "is_async" not in msg:
        msg["is_async"] = async_

    if info is not None:
        msg["info"] = info

    if script:
        msg["code"] = script.read()
    elif "code" not in msg:
        raise click.UsageError("Missing script")

    if "vars" in msg:
        vs = set(msg["vars"])
    else:
        vs = set()
    vd = msg.setdefault("default", {})

    for k, v in vars_:
        vs.add(k)
        vd[k] = v
    for k, v in eval_:
        vs.add(k)
        if v == "-":
            vd.pop(k, None)
        elif v == "/":
            vd.pop(k, None)
            vs.discard(k)
        else:
            vd[k] = eval(v)  # pylint:disable=eval-used
    for k, v in path_:
        vs.add(k)
        vd[k] = P(v)
    msg["vars"] = list(vs)

    res = await obj.client.set(
        obj.cfg["codes"]["prefix"] + path,
        value=msg,
        nchain=obj.meta,
        **({"chain": chain} if chain is not NotGiven else {}),
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.group("module")
async def mod():
    """
    Change the code of a module stored in DistKV
    """


@mod.command("get")
@click.option("-s", "--script", type=click.File(mode="w", lazy=True), help="Save the code here")
@click.argument("path", nargs=1)
@click.pass_obj  # pylint: disable=function-redefined
async def get_mod(obj, path, script):
    """Read a module entry"""
    path = P(path)
    if not len(path):
        raise click.UsageError("You need a non-empty path.")
    res = await obj.client._request(
        action="get_value", path=obj.cfg["modules"]["prefix"] + path, iter=False, nchain=obj.meta
    )
    if not obj.meta:
        res = res.value

    code = res.pop("code", None)
    if code is not None:
        code = code.rstrip("\n \t") + "\n"
        if script:
            print(code, file=script)
        else:
            res["code"] = code

    yprint(res, stream=obj.stdout)


@mod.command("set")
@click.option("-s", "--script", type=click.File(mode="r"), help="File with the module's code")
@click.option("-d", "--data", type=click.File(mode="r"), help="load the metadata (YAML)")
@click.argument("path", nargs=1)  # pylint: disable=function-redefined
@click.pass_obj
async def set_mod(obj, path, script, data):
    """Save a Python module to DistKV."""
    path = P(path)
    if not len(path):
        raise click.UsageError("You need a non-empty path.")

    if data:
        msg = yload(data)
    else:
        msg = {}
    chain = None
    if "value" in msg:
        chain = msg.get("chain", None)
        msg = msg["value"]

    if "code" not in msg:
        if script:
            raise click.UsageError("Duplicate script")
    else:
        if not script:
            raise click.UsageError("Missing script")
        msg["code"] = script.read()

    res = await obj.client.set(
        *obj.cfg["modules"]["prefix"], *path, value=msg, iter=False, nchain=obj.meta, chain=chain
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command("list")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m", "--maxdepth", type=int, default=None, help="Limit recursion depth. Default: whole tree"
)
@click.option(
    "-M", "--mindepth", type=int, default=None, help="Starting depth. Default: whole tree"
)
@click.option("-f", "--full", is_flag=True, help="print complete entries.")
@click.option("-s", "--short", is_flag=True, help="print shortened entries.")
@click.argument("path", nargs=1)
@click.pass_obj
async def list_(obj, path, as_dict, maxdepth, mindepth, full, short):
    """
    List code entries.

    Be aware that the whole subtree will be read before anything is
    printed if you use the `--as-dict` option.
    """

    path = P(path)
    if (full or as_dict) and short:
        raise click.UsageError("'-f'/'-d' and '-s' are incompatible.")
    kw = {}
    if maxdepth is not None:
        kw["max_depth"] = maxdepth
    if mindepth is not None:
        kw["min_depth"] = mindepth
    y = {}
    async for r in obj.client.get_tree(obj.cfg["codes"].prefix + path, nchain=obj.meta, **kw):
        r.pop("seq", None)
        path = r.pop("path")
        if not full:
            if "info" not in r.value:
                r.value.info = f"<{len(r.value.code.splitlines())} lines>"
            del r.value["code"]

        if short:
            print(path, "::", r.value.info)
            continue

        if as_dict is not None:
            yy = y
            for p in path:
                yy = yy.setdefault(p, {})
            try:
                yy[as_dict] = r if obj.meta else r.value
            except AttributeError:
                continue
        else:
            y = {}
            try:
                y[path] = r if obj.meta else r.value
            except AttributeError:
                continue
            yprint([y], stream=obj.stdout)

    if as_dict is not None:
        yprint(y, stream=obj.stdout)
    return


@cli.command()
@click.argument("path", nargs=1)
@click.pass_obj
async def delete(obj, path):
    """Remove a code entry"""
    path = P(path)

    res = await obj.client.get(obj.cfg["codes"].prefix + path, nchain=3)
    if "value" not in res:
        res.info = "Does not exist."
    else:
        res = await obj.client.delete(obj.cfg["codes"].prefix + path, chain=res.chain)
        res.info = "Deleted."

    if obj.meta:
        yprint(res, stream=obj.stdout)
    elif obj.debug:
        print(res.info, file=obj.stdout)



import pytest
import trio
import anyio
from asyncscope import scope

from distkv.mock.mqtt import stdtest

from distkv.client import ServerError
from distkv.util import PathLongener, P
from functools import partial

import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_21_load_save(autojump_clock, tmpdir):  # pylint: disable=unused-argument
    """also used to check watching"""
    path = tmpdir.join("foo")
    logger.debug("START")
    msgs = []
    s = None

    async def watch_changes(c, evt):
        lg = PathLongener(())
        async with c.watch(P(":"), nchain=3, fetch=True) as res:
            await evt.set()
            async for m in res:
                logger.info(m)
                lg(m)
                if m.get("value", None) is not None:
                    msgs.append(m)

    async with stdtest(args={"init": 234}, tocks=30) as st:
        (s,) = st.s
        async with st.client() as c:
            assert (await c.get(P(":"))).value == 234
            evt = anyio.create_event()
            cs = await scope.spawn(watch_changes, c, evt)
            await evt.wait()

            await c.set(P("foo"), value="hello", nchain=3)
            await c.set(P("foo.bar"), value="baz", nchain=3)
            await c.set(P(":"), value=2345, nchain=3)
            await trio.sleep(1)  # allow the writer to write
            await cs.cancel()
            pass  # client end

        logger.debug("SAVE %s", path)
        await s.save(path)
        logger.debug("SAVED")
        pass  # server end

    logger.debug("NEXT")
    for m in msgs:
        m.pop("tock", None)
        m.pop("seq", None)
    assert sorted(msgs, key=lambda x: x.chain.tick if "chain" in x else 0) == [
        {"chain": {"node": "test_0", "prev": None, "tick": 1}, "path": P(":"), "value": 234},
        {"chain": {"node": "test_0", "prev": None, "tick": 2}, "path": P("foo"), "value": "hello"},
        {
            "chain": {"node": "test_0", "prev": None, "tick": 3},
            "path": P("foo.bar"),
            "value": "baz",
        },
        {"chain": {"node": "test_0", "prev": None, "tick": 4}, "path": P(":"), "value": 2345},
    ]

    msgs = []
    async with stdtest(run=False, tocks=40) as st:
        (s,) = st.s
        logger.debug("LOAD %s", path)
        await s.load(path, local=True)
        logger.debug("LOADED")

        evt = anyio.create_event()
        await st.tg.spawn(partial(st.s[0].serve, ready_evt=evt))
        await evt.wait()

        logger.debug("RUNNING")

        async with st.client() as c:
            evt = anyio.create_event()
            cs = await scope.spawn(watch_changes, c, evt)
            await evt.wait()

            await c.set(P("foof"), value="again")
            assert (await c.get(P("foo"))).value == "hello"
            assert (await c.get(P("foo.bar"))).value == "baz"
            assert (await c.get(P(":"))).value == 2345
            await trio.sleep(1)  # allow the writer to write
            await cs.cancel()

    for m in msgs:
        m.pop("tock", None)
        m.pop("seq", None)
    assert sorted(msgs, key=lambda x: x.chain.tick if "chain" in x else 0) == [
        {"chain": {"node": "test_0", "prev": None, "tick": 2}, "path": P("foo"), "value": "hello"},
        {
            "chain": {"node": "test_0", "prev": None, "tick": 3},
            "path": P("foo.bar"),
            "value": "baz",
        },
        {"chain": {"node": "test_0", "prev": None, "tick": 4}, "path": P(":"), "value": 2345},
        {
            "chain": {"node": "test_0", "prev": None, "tick": 5},
            "path": P("foof"),
            "value": "again",
        },
    ]
    logger.debug("OK")


@pytest.mark.trio
async def test_02_cmd(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=50) as st:
        async with st.client() as c:
            assert (await c.get(P(":"))).value == 123
            r = await st.run("data set -v hello foo")
            r = await st.run("data set -ev 'baz' foo.bar")

            r = await st.run("data get :")
            assert r.stdout == "123\n"

            r = await st.run("data get foo")
            assert r.stdout == "'hello'\n"

            r = await st.run("data get foo.bar")
            assert r.stdout == "'baz'\n"

            r = await c._request(
                "get_state",
                nodes=True,
                known=True,
                missing=True,
                remote_missing=True,
                present=True,
            )
            del r["tock"]
            del r["seq"]
            assert r == {
                "node": "test_0",
                "nodes": {"test_0": 3},
                "known": {},
                "present": {"test_0": ((1, 4),)},
                "missing": {},
                "remote_missing": {},
            }

            assert (await c._request("get_value", node="test_0", tick=1)).value == 123
            assert (await c._request("get_value", node="test_0", tick=2)).value == "hello"
            assert (await c._request("get_value", node="test_0", tick=3)).value == "baz"

            r = await c.set(P(":"), value=1234, nchain=3)
            assert r.prev == 123
            assert r.chain.tick == 4

            # does not yet exist
            with pytest.raises(ServerError):
                await c._request("get_value", node="test_0", tick=8)
            # has been superseded
            with pytest.raises(ServerError):
                await c._request("get_value", node="test_0", tick=1)
            # works
            assert (await c._request("get_value", node="test_0", tick=4)).value == 1234

            r = await c._request(
                "get_state",
                nodes=True,
                known=True,
                missing=True,
                remote_missing=True,
                present=True,
            )
            del r["tock"]
            del r["seq"]
            assert r == {
                "node": "test_0",
                "nodes": {"test_0": 4},
                "known": {"test_0": (1,)},
                "present": {"test_0": ((2, 5),)},
                "missing": {},
                "remote_missing": {},
            }
            pass  # client end
        pass  # server end


@pytest.mark.trio
async def test_03_three(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(test_1={"init": 125}, n=2, tocks=30) as st:
        async with st.client(1) as ci:
            assert (await ci.get(P(":"))).value == 125

            r = await ci._request(
                "get_state",
                nodes=True,
                known=True,
                missing=True,
                remote_missing=True,
                present=True,
            )
            del r["tock"]
            del r["seq"]
            # Various stages of integrating test_0
            assert (
                r
                == {
                    "node": "test_1",
                    "nodes": {"test_1": 1},
                    "known": {},
                    "present": {"test_1": (1,)},
                    "missing": {},
                    "remote_missing": {},
                }
                or r
                == {
                    "node": "test_1",
                    "nodes": {"test_0": None, "test_1": 1},
                    "known": {},
                    "present": {"test_1": (1,)},
                    "missing": {},
                    "remote_missing": {},
                }
                or r
                == {
                    "node": "test_1",
                    "nodes": {"test_0": None, "test_1": 1},
                    "known": {},
                    "present": {"test_1": (1,)},
                    "missing": {"test_0": (1,)},
                    "remote_missing": {"test_0": (1,)},
                }
                or r
                == {
                    "node": "test_1",
                    "nodes": {"test_1": 1, "test_0": None},
                    "known": {},
                    "present": {"test_0": (1,), "test_1": (1,)},
                    "missing": {},
                    "remote_missing": {},
                }
                or r
                == {
                    "node": "test_1",
                    "nodes": {"test_0": 0, "test_1": 1},
                    "known": {},
                    "present": {"test_1": (1,)},
                    "missing": {},
                    "remote_missing": {},
                }
                or False
            )

            # This waits for test_0 to be fully up and running.
            async with st.client(0) as c:

                # At this point ci shall be fully integrated, and test_1 shall know this (mostly).
                r = await ci._request(
                    "get_state",
                    nodes=True,
                    known=True,
                    present=True,
                    missing=True,
                    remote_missing=True,
                )
                del r["tock"]
                del r["seq"]
                assert r == {
                    "node": "test_1",
                    "nodes": {"test_0": 0, "test_1": 1},
                    "known": {},
                    "present": {"test_1": (1,)},
                    "missing": {},
                    "remote_missing": {},
                } or r == {
                    "node": "test_1",
                    "nodes": {"test_0": None, "test_1": 1},
                    "known": {},
                    "present": {"test_1": (1,)},
                    "missing": {},
                    "remote_missing": {},
                }

                assert (await c.get(P(":"))).value == 125

                r = await c.set(P(":"), value=126, nchain=3)
                assert r.prev == 125
                assert r.chain.tick == 1
                assert r.chain.node == "test_0"
                assert r.chain.prev.tick == 1
                assert r.chain.prev.node == "test_1"
                assert r.chain.prev.tick == 1

                # This verifies that the chain entry for the initial update is gone
                # and the initial change is no longer retrievable.
                # We need the latter to ensure that there are no memory leaks.
                await trio.sleep(1)
                r = await ci.set(P(":"), value=127, nchain=3)
                assert r.prev == 126
                assert r.chain.tick == 2
                assert r.chain.node == "test_1"
                assert r.chain.prev.tick == 1
                assert r.chain.prev.node == "test_0"
                assert r.chain.prev.tick == 1
                assert r.chain.prev.prev is None
                await trio.sleep(1)

                with pytest.raises(ServerError):
                    await c._request("get_value", node="test_1", tick=1)
                with pytest.raises(ServerError):
                    await ci._request("get_value", node="test_1", tick=1)

                # Now test that the internal states match.
                await trio.sleep(1)
                r = await c._request(
                    "get_state",
                    nodes=True,
                    known=True,
                    present=True,
                    missing=True,
                    remote_missing=True,
                )
                del r["tock"]
                del r["seq"]
                assert r == {
                    "node": "test_0",
                    "nodes": {"test_0": 1, "test_1": 2},
                    "known": {"test_1": (1,)},
                    "present": {"test_0": (1,), "test_1": (2,)},
                    "missing": {},
                    "remote_missing": {},
                }
                pass  # client2 end

            r = await ci._request(
                "get_state",
                nodes=True,
                known=True,
                missing=True,
                remote_missing=True,
                present=True,
            )
            del r["tock"]
            del r["seq"]
            assert r == {
                "node": "test_1",
                "nodes": {"test_0": 1, "test_1": 2},
                "known": {"test_1": (1,)},
                "present": {"test_0": (1,), "test_1": (2,)},
                "missing": {},
                "remote_missing": {},
            }
            pass  # client end
        pass  # server end



"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.
"""

import msgpack
from functools import partial

from .util import attrdict, Path

__all__ = ["packer", "unpacker", "stream_unpacker"]


def _encode(data):
    if isinstance(data, int) and data >= 1 << 64:
        return msgpack.ExtType(2, data.to_bytes((data.bit_length() + 7) // 8, "big"))
    elif isinstance(data, Path):
        return msgpack.ExtType(3, b"".join(packer(x) for x in data))
    return data


def _decode(code, data):
    if code == 2:
        return int.from_bytes(data, "big")
    elif code == 3:
        s = stream_unpacker()
        s.feed(data)
        return Path(*s)
    return msgpack.ExtType(code, data)


# single message packer
_packers = []


def packer(data):
    if _packers:
        pack = _packers.pop()
    else:
        pack = msgpack.Packer(strict_types=False, use_bin_type=True, default=_encode).pack
    try:
        return pack(data)
    finally:
        _packers.append(pack)


# single message unpacker
unpacker = partial(
    msgpack.unpackb, object_pairs_hook=attrdict, raw=False, use_list=False, ext_hook=_decode
)

# stream unpacker factory
stream_unpacker = partial(
    msgpack.Unpacker, object_pairs_hook=attrdict, raw=False, use_list=False, ext_hook=_decode
)



import pytest

from distkv.mock.mqtt import stdtest

import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_51_dh(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}) as st:
        (s,) = st.s
        async with st.client() as c:
            assert len(s._clients) == 1
            sc = list(s._clients)[0]
            assert c._dh_key is None
            assert sc._dh_key is None
            dh = await c.dh_secret(length=10)
            assert dh == c._dh_key
            assert dh == sc._dh_key



import pytest
import trio

from distkv.mock.mqtt import stdtest

from distkv.client import ServerError
from distkv.util import P

import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_11_kill(autojump_clock):  # pylint: disable=unused-argument
    """also used to check watching"""
    async with stdtest(args={"init": 234}, n=3, tocks=2000) as st:
        async with st.client(1) as c:

            await c.set(P("foo"), value="hello")
            await c.set(P("foo.bar"), value="baz")
        await trio.sleep(1)

        async with st.client(2) as c:
            await c.set(P("foo"), value="super")
            await c.set(P("foo.bar"), value="seded")
        await trio.sleep(1)

        async with st.client(0) as c:
            res = await c.get(P("foo.bar"), nchain=3)
            assert res.chain == dict(
                node="test_2", tick=2, prev=dict(node="test_1", tick=2, prev=None)
            )

            res = await c._request("enum_node", node="test_1", current=True)
            assert res.result == ()

            res = await c._request("enum_node", node="test_1", current=False)
            assert res.result == (1, 2)

            res = await c._request("enum_node", node="test_2", current=True)
            assert res.result == (1, 2)

            # error
            with pytest.raises(ServerError):
                res = await c._request("kill_node", node="test_2", iter=False)

            # works
            res = await c._request("kill_node", node="test_1")
            await trio.sleep(2)

            res = await c.get(P("foo.bar"), nchain=3)
            assert res.chain == dict(node="test_2", tick=2, prev=None)



# from asyncclick.testing import CliRunner
import io
import sys
import attr
import socket

from distkv.client import open_client
from distkv.command import main
from distkv.default import CFG

try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager


async def run(*args, expect_exit=0, do_stdout=True):
    args = ("-c", "/dev/null", *args)
    if do_stdout:
        CFG["_stdout"] = out = io.StringIO()
    try:
        print("*****", args)
        res = await main.main(args)
        return res
    except SystemExit as exc:
        res = exc
        assert exc.code == expect_exit, exc.code
        return exc
    except BaseException as exc:
        res = exc
        raise
    else:
        assert expect_exit == 0
        return res
    finally:
        if do_stdout:
            res.stdout = out.getvalue()
            CFG["_stdout"] = sys.stdout


@attr.s
class S:
    tg = attr.ib()
    client_ctx = attr.ib()
    serfs = attr.ib(factory=set)
    s = attr.ib(factory=list)  # servers
    c = attr.ib(factory=list)  # clients
    _seq = 1

    async def ready(self, i=None):
        if i is not None:
            await self.s[i].is_ready
            return self.s[i]
        for s in self.s:
            if s is not None:
                await s.is_ready
        return self.s

    def __iter__(self):
        return iter(self.s)

    @asynccontextmanager
    async def client(self, i: int = 0, **kv):
        """Get a client for the i'th server."""
        await self.s[i].is_serving
        self._seq += 1
        for host, port, *_ in self.s[i].ports:
            if host[0] == ":":
                continue
            try:
                async with open_client(
                    _main_name="_client_%d_%d" % (i, self._seq),
                    connect=dict(host=host, port=port, ssl=self.client_ctx, **kv),
                ) as c:
                    yield c
                    return
            except socket.gaierror:
                pass
        raise RuntimeError("Duh? no connection")

    async def run(self, *args, do_stdout=True):
        h = p = None
        for s in self.s:
            for h, p, *_ in s.ports:
                if h[0] != ":":
                    break
            else:
                continue
            break
        if len(args) == 1:
            args = args[0]
            if isinstance(args, str):
                args = args.split(" ")
        return await run("client", "-h", h, "-p", p, *args, do_stdout=do_stdout)



import anyio
from distmqtt.client import MQTTClient
from distmqtt.codecs import NoopCodec
from contextlib import asynccontextmanager
from . import Backend
import logging

logger = logging.getLogger(__name__)

# Simply setting connect=asyncserf.serf_client interferes with mocking
# when testing.


class MqttMessage:
    def __init__(self, topic, payload):
        self.topic = topic
        self.payload = payload


class MqttBackend(Backend):
    client = None

    @asynccontextmanager
    async def connect(self, *a, **kw):
        C = MQTTClient(self._tg, codec=NoopCodec())
        try:
            await C.connect(*a, **kw)
            self.client = C
            yield self
        finally:
            self.client = None
            async with anyio.open_cancel_scope(shield=True):
                await C.disconnect()

    @asynccontextmanager
    async def monitor(self, *topic):
        topic = "/".join(topic)
        logger.info("Monitor %s start", topic)
        try:
            async with self.client.subscription(topic) as sub:

                async def sub_get(sub):
                    async for msg in sub:
                        yield MqttMessage(msg.topic.split("/"), msg.data)

                yield sub_get(sub)
        except anyio.get_cancelled_exc_class():
            raise
        except BaseException as exc:
            logger.exception("Monitor %s end: %r", topic, exc)
            raise
        else:
            logger.info("Monitor %s end", topic)

    def send(self, *topic, payload):  # pylint: disable=invalid-overridden-method
        """
        Send this payload to this topic.
        """
        # client.publish is also async, pass-thru
        return self.client.publish("/".join(topic), message=payload)


@asynccontextmanager
async def connect(**kw):
    async with anyio.create_task_group() as tg:
        c = MqttBackend(tg)
        async with c.connect(**kw):
            yield c



import pytest

from distkv.mock.mqtt import stdtest
from distkv.client import ServerError
from distkv.util import PathLongener, P

import logging

logger = logging.getLogger(__name__)

pytestmark = pytest.mark.xfail()


async def collect(i, path=()):
    res = []
    pl = PathLongener(path)
    async for r in i:
        r.pop("tock", 0)
        r.pop("seq", 0)
        pl(r)
        res.append(r)
    return res


@pytest.mark.trio
async def test_41_ssl_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(ssl=True, args={"init": 123}) as st:
        async with st.client() as c:
            assert (await c.get()).value == 123

            r = await c.set(P("foo"), value="hello", nchain=3)
            r = await c.set(P("foo.bar"), value="baz", nchain=3)
            r = await c.get()
            assert r.value == 123

            r = await c.get("foo")
            assert r.value == "hello"

            exp = [
                {"path": P(":"), "value": 123},
                {"path": P("foo"), "value": "hello"},
                {"path": P("foo.bar"), "value": "baz"},
            ]
            async with c._stream("get_tree", path=P(":"), max_depth=2) as rr:
                r = await collect(rr)
            assert r == exp

            exp.pop()
            async with c._stream("get_tree", path=P(":"), iter=True, max_depth=1) as rr:
                r = await collect(rr)
            assert r == exp

            exp.pop()
            async with c._stream("get_tree", path=P(":"), iter=True, max_depth=0) as rr:
                r = await collect(rr)
            assert r == exp

            r = await c.get(P("foo.bar"))
            assert r.value == "baz"

            r = await c._request(
                "get_state",
                nodes=True,
                known=True,
                missing=True,
                remote_missing=True,
                present=True,
            )
            del r["tock"]
            del r["seq"]
            assert r == {
                "node": "test_0",
                "nodes": {"test_0": 3},
                "known": {},
                "present": {"test_0": ((1, 4),)},
                "missing": {},
                "remote_missing": {},
            }

            assert (await c._request("get_value", node="test_0", tick=1)).value == 123
            assert (await c._request("get_value", node="test_0", tick=2)).value == "hello"
            assert (await c._request("get_value", node="test_0", tick=3)).value == "baz"

            r = await c.set(value=1234, nchain=3)
            assert r.prev == 123
            assert r.chain.tick == 4

            # does not yet exist
            with pytest.raises(ServerError):
                await c._request("get_value", node="test_0", tick=8)
            # has been superseded
            with pytest.raises(ServerError):
                await c._request("get_value", node="test_0", tick=1)
            # works
            assert (await c._request("get_value", node="test_0", tick=4)).value == 1234

            r = await c._request(
                "get_state",
                nodes=True,
                known=True,
                missing=True,
                remote_missing=True,
                present=True,
            )
            del r["tock"]
            del r["seq"]
            assert r == {
                "node": "test_0",
                "nodes": {"test_0": 4},
                "known": {"test_0": (1,)},
                "present": {"test_0": ((2, 5),)},
                "missing": {},
                "remote_missing": {},
            }
            pass  # client end
        pass  # server end



# distkv.auth
# template for authorization

"""
This set of modules authenticates users.

A submodule is expected to export a "load(type:str, make:bool, server:bool)"
method that returns a class.

It must recognize, at minimum:

* load("stream", server:bool):

  A filter used for logging in a user.

* load("user", server:bool, make:bool)

  A class used to represent the user, or a way to create/manipulate a user record.

The client process is:

* create a user:

  * Create a :class:`BaseUserMaker` by calling :meth:`BaseUserMaker.build`
    with a record conforming to its schema.

  * Export that and save it to the server, at (None,"auth","user",NAME).

* modify a user:

  * Call :class:`BaseUserMaker.import` with the value from the server.

* Log in:

  * Create a :class:`BaseUser` by calling :meth:`BaseUserMaker.build`
    with a record conforming to its schema.

  * Call :meth:`BaseUser.auth`.

The server process is:

* create a user:

  * The server intercepts the write call and uses the data to call
    :meth:`BaseServerAuthMaker.build`.

  * It calls :meth:`BaseServerMaker.save` and stores the actual result.

* modify a user:

  * the server calls :meth:`BaseServerAuth.read` with the stored data,
    then sends a record created with :meth:`BaseServerAuth.save` to the client.

* verify a user:

  * The server calls :meth:`BaseServerAuth.read` with the stored data.

  * The server calls :meth:`BaseServerAuth.auth` with the record from the client.

"""

import jsonschema
import io
from importlib import import_module
from ..client import NoData, Client
from ..model import Entry
from ..server import StreamCommand, ServerClient
from ..util import split_one, attrdict, NotGiven, yload, Path
from ..exceptions import NoAuthModuleError
from ..types import ACLFinder, NullACL

# Empty schema
null_schema = {"type": "object", "additionalProperties": False}

# Additional schema data for specific types
add_schema = {
    "user": {
        "acl": {
            "type": "object",
            "additionalProperties": False,
            "properties": {"key": {type: "string", "minLength": 1}},
        },
        "conv": {
            "type": "object",
            "additionalProperties": False,
            "properties": {"key": {type: "string", "minLength": 1}},
        },
    }
}


def loader(method: str, typ: str, *a, **k):
    m = method
    if "." not in m:
        m = "distkv.auth." + m
    cls = import_module(m).load(typ, *a, **k)
    cls._auth_method = method
    cls._auth_typ = typ
    cls.aux_schemas = add_schema.get(typ, null_schema)
    return cls


def gen_auth(s: str):
    """
    Generate auth data from parameters or YAML file (if first char is '=').
    """
    if not isinstance(s, str):
        return s  # called twice. Oh well.

    m, *p = s.split()
    if len(p) == 0 and m[0] == "=":
        with io.open(m[1:], "r") as f:
            kw = yload(f)
            m = kw.pop("type")
    else:
        kw = {}
        for pp in p:
            split_one(pp, kw)
    try:
        m = loader(m, "user", server=False)
    except ModuleNotFoundError:
        raise NoAuthModuleError(m) from None
    return m.build(kw)


def load(typ: str, *, make: bool = False, server: bool):
    """
    This procedure is used to load and return a user management class.

    Arguments:
        typ: the type of module to load.
        make: flag that the caller wants a record-generating, not a
              record-using class.
        server: flag that the class is to be used on the server, not on the
                client.

    Types:
        stream: the filter used for authorizing the user.
        user: represents a user record.
    """
    raise NotImplementedError("You need to implement me")  # pragma: no cover


async def null_server_login(stream):
    return stream


async def null_client_login(stream, user: "BaseClientAuth"):  # pylint: disable=unused-argument
    return stream


def _load_example(typ: str, make: bool, server: bool):  # pragma: no cover
    """example code for :proc:`load`
    """
    if typ == "client":
        if server:
            return null_server_login
        else:
            return null_client_login
    if typ != "user":
        raise NotImplementedError("This module only handles users")
    if server:
        if make:
            return BaseServerAuthMaker
        else:
            return BaseServerAuth
    else:
        if make:
            return BaseClientAuthMaker
        else:
            return BaseClientAuth


class _AuthLoaded:
    # This class is mainly there to appease pylint
    _auth_method = None


class BaseClientAuth(_AuthLoaded):
    """
    This class is used for creating a data record which authenticates a user.

    The schema verifies the input to :meth:`build`.
    """

    schema = null_schema

    def __init__(self, **data):
        jsonschema.validate(instance=data, schema=type(self).schema)
        for k, v in data.items():
            setattr(self, k, v)

    @classmethod
    def build(cls, user):
        """
        Create a user record from the data conforming to this schema.
        """
        return cls(**user)

    @property
    def ident(self):
        """Some user identifier.
        Required so that the server can actually find the record.
        """
        return "*"

    async def auth(self, client: Client, chroot=()):
        """
        Authorizes this record with the server.
        """
        try:
            await client._request(
                action="auth",
                typ=self._auth_method,
                iter=False,
                chroot=chroot,
                ident=self.ident,
                data=self.auth_data(),
            )
        except NoData:
            pass

    def auth_data(self):
        """
        Additional data for the initial auth message.

        Does NOT include 'ident', that gets added explicitly by :meth:`auth`.
        """
        return {}


class BaseClientAuthMaker(_AuthLoaded):
    """
    This class is used for creating a data record which describes a user record.

    While :class:`BaseClientAuth` is used solely for authentication,
    this class is used to represent the server's user data.

    The schema verifies the input to :meth:`build`.
    """

    gen_schema = null_schema
    mod_schema = null_schema
    _chain = None

    def __init__(self, _initial=True, **data):
        if _initial:
            jsonschema.validate(instance=data, schema=type(self).gen_schema)
        else:
            jsonschema.validate(instance=data, schema=type(self).mod_schema)
        for k, v in data.items():
            setattr(self, k, v)

    @classmethod
    def build(cls, user, _initial=True):
        """
        Create a user record from the data conforming to this schema.
        """
        return cls(**user, _initial=_initial)

    def export(self):
        """Return the data required to re-create the user via :meth:`build`."""
        return {}  # pragma: no cover

    @property
    def ident(self):
        """The identifier for this user.

        Required so that the server can actually find the record.
        """
        return "*"  # pragma: no cover

    @classmethod
    async def recv(cls, client: Client, ident: str, _kind="user", _initial=True):
        """Read this user from the server.

        Sample code …
        """
        # pragma: no cover
        res = await client._request(
            "auth_get", typ=cls._auth_method, kind=_kind, ident=ident, nchain=0 if _initial else 2
        )
        self = cls(_initial=_initial)
        self._chain = res.chain
        return self

    async def send(self, client: Client, _kind="user"):
        """Send this user to the server."""
        try:
            await client._request(
                "auth_set",
                iter=False,
                typ=type(self)._auth_method,
                kind=_kind,
                ident=self.ident,
                chain=self._chain,
                data=self.send_data(),
            )
        except NoData:
            pass

    def send_data(self):
        return {}


class BaseServerAuth(_AuthLoaded):
    """
    This class is used on the server to represent / verify a user.

    The schema verifies whatever data the associated ``ClientAuth`` initially sends.
    """

    schema = null_schema.copy()
    schema["additionalProperties"] = True

    is_super_root = False
    can_create_subtree = False
    can_auth_read = False
    can_auth_write = False

    def __init__(self, data: dict = {}):  # pylint: disable=dangerous-default-value
        if data:
            for k, v in data.items():
                setattr(self, k, v)

    @classmethod
    def load(cls, data: Entry):
        """Create a ServerAuth object from existing stored data"""
        return cls(data.data)

    async def auth(self, cmd: StreamCommand, data):  # pylint: disable=unused-argument
        """Verify that @data authenticates this user."""
        jsonschema.validate(instance=data.get("data", {}), schema=type(self).schema)

    def aux_conv(self, data: Entry, root: Entry):
        from ..types import ConvNull

        try:
            data = data["conv"].data["key"]
            res, _ = root.follow_acl(Path(None, "conv", data), create=False, nulls_ok=True)
            return res
        except (KeyError, AttributeError):
            return ConvNull

    def aux_acl(self, data: Entry, root: Entry):
        try:
            data = data["acl"].data["key"]
            if data == "*":
                return NullACL
            acl, _ = root.follow_acl(Path(None, "acl", data), create=False, nulls_ok=True)
            return ACLFinder(acl)
        except (KeyError, AttributeError):
            return NullACL

    def info(self):
        """
        Return whatever public data the user might want to have displayed.

        This includes information to identify the user, but not anything
        that'd be suitable for verifying or even faking authorization.
        """
        return {}

    async def check_read(
        self, *path, client: ServerClient, data=None
    ):  # pylint: disable=unused-argument
        """Check that this user may read the element at this location.
        This method may modify the data.
        """
        return data

    async def check_write(
        self, *path, client: ServerClient, data=None
    ):  # pylint: disable=unused-argument
        """Check that this user may write the element at this location.
        This method may modify the data.
        """
        return data


class RootServerUser(BaseServerAuth):
    """The default user when no auth is required

    Interim record. TODO: create a separate ACL thing.
    """

    is_super_root = True
    can_create_subtree = True
    can_auth_read = True
    can_auth_write = True


class BaseServerAuthMaker(_AuthLoaded):
    """
    This class is used on the server to verify the transmitted user record
    and to store it in DistKV.

    The schema verifies the data from the client.
    """

    schema = null_schema
    aux_schemas = None  # set by the loader

    def __init__(self, chain=None, data=None):
        if data is not None and data is not NotGiven:
            for k, v in data.items():
                setattr(self, k, v)
        self._chain = chain

    @classmethod
    def load(cls, data: Entry):
        """Read the user data from DistKV"""
        return cls(chain=data.chain, data=data.data)

    @classmethod
    async def recv(
        cls, cmd: StreamCommand, data: attrdict  # pylint: disable=unused-argument
    ) -> "BaseServerAuthMaker":
        """Create/update a new user by reading the record from the client"""
        dt = data.get("data", None) or {}
        jsonschema.validate(instance=dt, schema=cls.schema)
        self = cls(chain=data["chain"], data=dt)
        return self

    @property
    def ident(self):
        """The record to store this user under."""
        return "*"

    def save(self):
        """Return a record to represent this user, suitable for saving to DistKV"""
        # does NOT contain "ident" or "chain"!
        return {}

    async def send(self, cmd: StreamCommand):  # pylint: disable=unused-argument
        """Send a record to the client, possibly multi-step / secured / whatever"""
        res = {}
        res["chain"] = self._chain.serialize() if self._chain else None
        return res
